
#############################################################################
#                                                                           #
#                       Offline Installer version 1.0                       #
#                           So far Windows only                             #
#                                                                           #
#############################################################################

# Getting the name of the temporary directory with installation files
temp.histfile <- tempfile("Rrawhist")
savehistory(temp.histfile )
rawhist <- readLines(temp.histfile)
ee<-rawhist[length(rawhist)]

ee<-gsub("source[(]\"","",ee)
ee<-gsub("\"[)]","",ee)
ee<-gsub("\\","/",ee,fixed=TRUE)
ee<-gsub("/{1,}","/",ee)

installDir<-dirname(ee)
unlink(temp.histfile)
setwd(installDir)

# Get all the root library names available in the temporary installation directory
lib.lst<-list.files(pattern="zip$")
lib.names<-gsub("([-_.0-9]{1,})(zip)","",lib.lst)

# What we need
lib.install<-c("vioplot","sm","zoo") 

# Check whether all libraries needed are present in the temporary installation directory
eee<-!lib.install%in%lib.names
if(any(eee)){
    cat("Library(ies)",paste(lib.install[eee],collapse=", "),"not found!\n")
    cat("Expect problems...\n")
}else{
    cat("All libraries available for installation, good...\n")
}

# Get the installation list       
ee<-lib.names%in%lib.install
lib.todo<-lib.lst[ee]

# Run the installation
out<-sapply(lib.todo,function(i){
    print(i)
    install.packages(i,repos=NULL) 
},simplify=TRUE)

cat("Libraries:",paste(c("  ",names(out)),collapse="\n  "),"\nwere installed succesfully!\n\n",sep="") 

###############################################################################################################

# destination directory
options(warn=-1)
library("GCDkit")
options(show.error.messages=FALSE)
destDirName<-gcdx.dir
    
# Patching 6.2.0
if(installed.packages()["GCDkit","Version"]=="6.2.0"){
    source(paste(installDir,"installPatch240226.r",sep="/"))
    cat("GCDkit 6.2.0 Patched!\n",sep="")
    flush.console() 
}

# The plugin itself
ee<-file.copy(paste(installDir,"Hf.r",sep="/"),paste(destDirName,"Plugin","Hf.r",sep="/"), overwrite = TRUE)
ee<-file.copy(paste(installDir,"constants_hfIso.xxx",sep="/"),paste(destDirName,"Plugin","constants_hfIso.xxx",sep="/"), overwrite = TRUE)
cat("Plugin files uploaded!\n",sep="") 

# Sample datasets
ee<-file.copy(paste(installDir,"data","khantaishir_Hf.csv",sep="/"),paste(destDirName,"data","khantaishir_Hf.csv",sep="/"), overwrite = TRUE)
ee<-file.copy(paste(installDir,"data","tatric_Hf.csv",sep="/"),paste(destDirName,"data","tatric_Hf.csv",sep="/"), overwrite = TRUE)
cat("Sample dataset(s) uploaded!\n",sep="") 

# The html help
which.files<-list.files(paste(installDir,"html",sep="/"),full.names=TRUE)
ee<-file.copy(which.files,paste(destDirName,"html",sep="/"), overwrite = FALSE)

# Help figs
which.files<-list.files(paste(installDir,"help/figures",sep="/"),full.names=TRUE)
ee<-file.copy(which.files,paste(destDirName,"help/figures",sep="/"), overwrite = TRUE)

# Amend Plugins help system
# Append or rewrite HafAn html help
filek.in<-paste(destDirName,"GCDkit_help/Plugin/00index.html",sep="/")
prd<-scan(file=filek.in,what="character",sep="\n",quiet=TRUE)
hafan.pos<-grep("HafAn",prd)
html.end.pos<-grep("</html>",prd)

filek.chunk<-paste(installDir,"help/00indexHf.html",sep="/")
hafanhelp<-scan(file=filek.chunk,what="character",sep="\n",quiet=TRUE)

if(length(hafan.pos)==0){
    out<-c(prd[1:(html.end.pos-2)],"",hafanhelp,"  <br><br>","  </body>","</html>")
}else{
    hafan.end.pos<-grep("<br><br>",prd[hafan.pos:length(prd)])[1]
    out<-c(prd[1:(hafan.pos-1)],"",hafanhelp,prd[(hafan.pos+hafan.end.pos):length(prd)])
}
write(out,file = filek.in,sep="\n",append=FALSE)
cat("GCDkit help system updated!\n",sep="") 
