if(length(grep("^176Hf/177Hf",colnames(WR)))>0&length(grep("^176Lu/177Hf",colnames(WR)))>0){
library("lattice")

#############################################################################
#                                                                           #
#                Reading a configuration file with the constants            #
#                                                                           #
#############################################################################

# Saving the Hf isotopic parameters into the setup file
# The argument is a numeric vector of the length four: c(selectedLambda,selectedDM,selectedCHUR,selectedRCC)
.saveHfParams<-function(selectedHfPars){ 
    filename<-paste(gcdx.dir,"/Plugin/constants_hfIso.xxx",sep="")
    ee<-scan(filename,what="raw",encoding="utf8",comment.char="",sep="\n",quiet=TRUE)
    
    # Get sequence number of the line where definition of each parameter starts
    tits<-grep("^[a-zA-Z0-9]",ee,value=FALSE)
    omit<-grep("^Selected",ee)
    tits<-tits[-which(tits%in%omit)]
    tits<-c(tits,length(ee)+1)

    # Strip all old stars
    ee<-gsub("[*]","",ee)
    
    # Marking new stars
    prd<-lapply(2:length(tits),function(i){
        # Which text line is to be replaced?
        txt.line<-tits[i-1]+1+1+1+selectedHfPars[i-1] # TODO Count the hashmarks, here we assume always exactly one
        ee[txt.line]<<-paste("*",ee[txt.line],sep="")
        invisible(i)
    })
    write(ee, file = filename,append = FALSE, sep = "")
}

# Loading the Hf isotopic parameters setup file, 
# assigns global variables lambda, CHUR, DM, R.crust

.loadHfParams<-function(){
    filename<-paste(gcdx.dir,"/Plugin/constants_hfIso.xxx",sep="")
    ee<-scan(filename,what="raw",encoding="utf8",comment.char="#",sep="\n",quiet=TRUE)

    # Getting complete lists of options for TclTk editor
    tits<-grep("^[a-zA-Z0-9]",ee,value=FALSE)
    omit<-grep("^Selected",ee)
    tits<-tits[-which(tits%in%omit)]

    params<-sapply(tits,function(j){unlist(strsplit(ee[j],"\t"))[1]},simplify=TRUE)

    tits<-c(tits,length(ee)+1)
    z<-lapply(2:length(tits),function(i){
        names<-unlist(strsplit(ee[tits[i-1]+1],"\t"))
        values<-unlist(strsplit(ee[(tits[i-1]+2):(tits[i]-1)],"\t"))
        z<-matrix(values,ncol=5,byrow=TRUE)
        colnames(z)<-names
        return(z)
    })

    # List of all available options
    names(z)<-params

    # Assigning the respective variables to the .GlobalEnv

            # 176Lu decay constant
            defLambda<-which(z[[1]][,1]=="*")
            lambda<-matrix(as.numeric(z[[1]][defLambda,"lambda"]))
            colnames(lambda)<-"Hf" 
            rownames(lambda)<-z[[1]][defLambda,"Reference"]
            assign("lambda",lambda,.GlobalEnv)

            # DM - present-day Hf isotopic composition
            defDM<-which(z[[2]][,1]=="*")
            DM<-matrix(as.numeric(z[[2]][defDM,4:5]),nrow=1)
            colnames(DM)<-colnames(z[[2]])[4:5]
            rownames(DM)<-paste("DM"," (",z[[2]][defDM,"Reference"],")",sep="")
            assign("DM",DM,.GlobalEnv)

            # CHUR - present-day Hf isotopic composition
            defCHUR<-which(z[[3]][,1]=="*")
            CHUR<-matrix(as.numeric(z[[3]][defCHUR,4:5]),nrow=1)
            colnames(CHUR)<-colnames(z[[3]])[4:5]
            rownames(CHUR)<-paste("CHUR"," (",z[[3]][defCHUR,"Reference"],")",sep="")
            assign("CHUR",CHUR,.GlobalEnv)

            # 176Lu/177Hf ratio of the source is unknown and thus needs to be assumed
            defRCC<-which(z[[4]][,1]=="*")
            R.crust<-matrix(as.numeric(z[[4]][defRCC,"176Lu/177Hf"]))
            colnames(R.crust)<-"176Lu/177Hf"
            rownames(R.crust)<-paste(paste(z[[4]][defRCC,c("Reservoir","Reference")],collapse=" ("),")",sep="")
            assign("R.crust",R.crust,.GlobalEnv)
    invisible(z)
}


#############################################################################
#                                                                           #
#                          Here comes the TclTk editor                      #
#                                                                           #
#############################################################################

.loadHfParams()

hfOptions<-function(){  
    require(tcltk) || stop("tcltk support is absent")

    # Restore parameters from the configuration file
    out.orig<-.loadHfParams()

    fontHeading<<-tkfont.create(family="arial",size=9,weight="bold")
    fontSmall<<- tkfont.create(family="arial",size=7,weight="bold")
    
    # Initialize Tcl variables

    # Decay constant
    listLambda<-out.orig[[1]][,"Reference"]
    defLambda<-which(out.orig[[1]][,1]=="*")
    tkLambda<-tclVar(defLambda) 

    # Depleted Mantle composition
    listDM<-out.orig[[2]][,"Reference"]
    defDM<-which(out.orig[[2]][,1]=="*")
    tkDM<-tclVar(defDM) 

    # CHUR composition
    listCHUR<-out.orig[[3]][,"Reference"]
    defCHUR<-which(out.orig[[3]][,1]=="*")
    tkCHUR<-tclVar(defCHUR) 
    
    # RCC = 176Lu/177Hf ratio of the crustal source for the two-stage model ages
    listRCC<-apply(out.orig[[4]],1,function(j){
        z<-paste(j[c("Reservoir","Reference")],collapse=" (")
        z<-paste(z,")",sep="")
    })
    defRCC<-which(out.orig[[4]][,1]=="*")
    tkRCC<-tclVar(defRCC)
    txt.width<-max(nchar(listRCC))
            
    # Open Tcl/Tk window 
    tt<<-tktoplevel()
    tkconfigure(tt,padx=10,pady=5,width=100,height=100)
    # tkconfigure(tt,cursor="left_ptr")
    tkwm.title(tt,"Hf isotopic options")

    main.frame <<- tkframe(tt)
    tkfocus(tt)
    
    # Lambda list box  
    frame.lambda <- tkframe(tt, borderwidth=1, relief="groove")            
        frame.lambda.refs <- tkframe(frame.lambda, borderwidth=0)
            OnLeave0 <- function(){

            }
            tkconfigure(frame.lambda.refs,padx=5,pady=5)
            tkgrid(tklabel(frame.lambda.refs,text="\u00B9\u2077\u2076Lu decay constant",background="gray95",
                font=fontHeading),row=0,column=1)
            listbox0<-.tkListBoxMy(frame.lambda.refs,listLambda,tkLambda,text="",selectmode="single",
                height=length(listLambda),width=txt.width,preselect=defLambda-1,on.leave=OnLeave0,row=1,column=1) 
        tkgrid(frame.lambda.refs,sticky="w",row=1,column=1)

        frame.lambda.vals <- tkframe(frame.lambda, borderwidth=0)
            tkconfigure(frame.lambda.vals,padx=5,pady=5)
            tkgrid(tklabel(frame.lambda.refs,text="yr\u207B\u00B9",background="gray95"),row=0,column=2)
            txt<-paste(out.orig[[1]][,"lambda"],collapse="\n")
            tkgrid(tklabel(frame.lambda.refs,text=txt,background="white"),row=1,column=2,sticky="n")

            tkgrid(tklabel(frame.lambda.refs,text="\u00B1",background="gray95"),row=0,column=3)
            txt<-paste(out.orig[[1]][,"error"],collapse="\n")
            tkgrid(tklabel(frame.lambda.refs,text=txt,background="white"),row=1,column=3,sticky="n")

        tkgrid(frame.lambda.vals,sticky="w",row=1,column=2)
    tkgrid(frame.lambda,sticky="we",columnspan=3) 

    # DM list box  
    frame.DM <- tkframe(tt, borderwidth=1, relief="groove")            
        frame.DM.refs <- tkframe(frame.DM, borderwidth=1)
            OnLeave1 <- function(){

            }
            tkconfigure(frame.DM.refs,padx=5,pady=5)
            tkgrid(tklabel(frame.DM.refs,text=names(out.orig)[2],background="gray95",font=fontHeading),row=0,column=1)
            listbox1<-.tkListBoxMy(frame.DM.refs,listDM,tkDM,text="",selectmode="single",
                height=length(listDM),width=txt.width,preselect=defDM-1,on.leave=OnLeave1,row=1,column=1)    
        tkgrid(frame.DM.refs,sticky="w",row=2,column=1)

        frame.DM.vals <- tkframe(frame.DM, borderwidth=0)
            tkconfigure(frame.DM.vals,padx=5,pady=5)
               
            tkgrid(tklabel(frame.DM.refs,text="\u00B9\u2077\u2076Lu/\u00B9\u2077\u2077Hf",background="gray95"),
                row=0,column=2,sticky="e")
            tkgrid(tklabel(frame.DM.refs,text="\u00B9\u2077\u2076Hf/\u00B9\u2077\u2077Hf",background="gray95"),
                row=0,column=3,sticky="e")
 
            txt<-paste(out.orig[[2]][,"176Lu/177Hf"],collapse="\n")
            tkgrid(tklabel(frame.DM.refs,text=txt,background="white"),row=1,column=2,sticky="n")
                
            txt<-paste(out.orig[[2]][,"176Hf/177Hf"],collapse="\n")
            tkgrid(tklabel(frame.DM.refs,text=txt,background="white"),row=1,column=3,sticky="n")
        tkgrid(frame.DM.vals,sticky="w",row=1,column=1)
    tkgrid(frame.DM,sticky="we",columnspan=3) 

    # CHUR list box  
    frame.CHUR <- tkframe(tt, borderwidth=1, relief="groove")            
        frame.CHUR.refs <- tkframe(frame.CHUR, borderwidth=1)
            OnLeave2 <- function(){

            }
            tkconfigure(frame.CHUR.refs,padx=5,pady=5)
            tkgrid(tklabel(frame.CHUR.refs,text=names(out.orig)[3],background="gray95",font=fontHeading),row=0,column=1)
            listbox2<-.tkListBoxMy(frame.CHUR.refs,listCHUR,tkCHUR,text="",selectmode="single",
                height=length(listCHUR),width=txt.width,preselect=defCHUR-1,on.leave=OnLeave2,row=1,column=1)    
        tkgrid(frame.CHUR.refs,sticky="w",row=2,column=1)

        frame.CHUR.vals <- tkframe(frame.CHUR, borderwidth=0)
            tkconfigure(frame.CHUR.vals,padx=5,pady=5)
               
            tkgrid(tklabel(frame.CHUR.refs,text="\u00B9\u2077\u2076Lu/\u00B9\u2077\u2077Hf",background="gray95"),
                row=0,column=2,sticky="e")
            tkgrid(tklabel(frame.CHUR.refs,text="\u00B9\u2077\u2076Hf/\u00B9\u2077\u2077Hf",background="gray95"),
                row=0,column=3,sticky="e")
 
            txt<-paste(out.orig[[3]][,"176Lu/177Hf"],collapse="\n")
            tkgrid(tklabel(frame.CHUR.refs,text=txt,background="white"),row=1,column=2,sticky="n")
                
            txt<-paste(out.orig[[3]][,"176Hf/177Hf"],collapse="\n")
            tkgrid(tklabel(frame.CHUR.refs,text=txt,background="white"),row=1,column=3,sticky="n")
        tkgrid(frame.CHUR.vals,sticky="w",row=1,column=1)
    tkgrid(frame.CHUR,sticky="we",columnspan=3) 

    # RCC list box
    frame.RCC <- tkframe(tt, borderwidth=1, relief="groove")            
        frame.RCC.refs <- tkframe(frame.RCC, borderwidth=1)
            OnLeave3 <- function(){

            }
            tkconfigure(frame.RCC.refs,padx=5,pady=5)
            tkgrid(tklabel(frame.RCC.refs,text="First stage of the two-stage Hf evolution models",background="gray95",font=fontHeading),row=0,column=1)
            listbox3<-.tkListBoxMy(frame.RCC.refs,listRCC,tkRCC,text="",selectmode="single",
                height=length(listRCC),width=txt.width,preselect=defRCC-1,on.leave=OnLeave3,row=1,column=1)    
        tkgrid(frame.RCC.refs,sticky="w",row=2,column=1)

        frame.RCC.vals <- tkframe(frame.RCC, borderwidth=0)
            tkconfigure(frame.RCC.vals,padx=5,pady=5)
               
            tkgrid(tklabel(frame.RCC.refs,text="\u00B9\u2077\u2076Lu/\u00B9\u2077\u2077Hf",background="gray95"),row=0,column=2,sticky="e")
 
            txt<-paste(out.orig[[4]][,"176Lu/177Hf"],collapse="\n")
            tkgrid(tklabel(frame.RCC.refs,text=txt,background="white"),row=1,column=2,sticky="n")
            tkgrid(frame.RCC.vals,sticky="w",row=1,column=1)
    tkgrid(frame.RCC,sticky="we",columnspan=3) 

###############
    
    # Save button  
    OnSave <- function(){
            # 176Lu decay constant
            ee0<-as.numeric(tkcurselection(listbox0))
            
            lambda<-matrix(as.numeric(out.orig[[1]][ee0+1,"lambda"]),nrow=1)
            colnames(lambda)<-"Hf" 
            rownames(lambda)<-listLambda[ee0+1]
            assign("lambda",lambda,.GlobalEnv)

            # DM - present-day Hf isotopic composition
            ee1<-as.numeric(tkcurselection(listbox1))
           
            DM<-matrix(as.numeric(out.orig[[2]][ee1+1,4:5]),nrow=1)
            colnames(DM)<-colnames(out.orig[[2]])[4:5]
            rownames(DM)<-paste("DM"," (",listDM[ee1+1],")",sep="")
            assign("DM",DM,.GlobalEnv)

            # CHUR - present-day Hf isotopic composition
            ee2<-as.numeric(tkcurselection(listbox2))

            CHUR<-matrix(as.numeric(out.orig[[3]][ee2+1,4:5]),nrow=1)
            colnames(CHUR)<-colnames(out.orig[[3]])[4:5]
            rownames(CHUR)<-paste("CHUR"," (",listCHUR[ee2+1],")",sep="")
            assign("CHUR",CHUR,.GlobalEnv)

            # 176Lu/177Hf ratio of the crustal source for the two-stage model ages
            ee3<-as.numeric(tkcurselection(listbox3))
            
            R.crust<-as.numeric(out.orig[[4]][ee3+1,"176Lu/177Hf"])
            names(R.crust)<-listRCC[ee3+1]
            assign("R.crust",R.crust,.GlobalEnv)
            
            # Saving Params INTO FILE
            .saveHfParams(c(ee0,ee1,ee2,ee3))
            .tkmessageBoxMy(type="ok",message="Saving successful") 
            tkdestroy(tt)
            
            # Recalculate
            hfIso(GUI=FALSE)
    }
    
    # Cancel button  
    OnCancel <- function(){
        cat("Cancelled....\n")
        flush.console()
        tkdestroy(tt)
    }
    
    # Restore defaults button
    OnDef<- function(){
        x<-winDialog("Are you sure?",type="yesno")
        if(x=="NO") return()
        .loadHfParams()

        flush.console()
        tkdestroy(tt)
        hfOptions()
    }
    
    # Help button
    OnHelp<-function(){
        browseURL(paste("file:/",gcdx.dir,"/html/hfOptions.html",sep="")) 
    }
    
    frame.but<- tkframe(tt, borderwidth=0)
        Save.but <-tkbutton(frame.but,text="   Ok/Save   ",command=OnSave,fg="darkgreen")
        Cancel.but <-tkbutton(frame.but,text="   Cancel   ",command=OnCancel,fg="darkred")
        Defaults.but<-tkbutton(frame.but,text="   Restore defaults   ",command=OnDef)
        Help.but <-tkbutton(frame.but,text="   Help   ",command=OnHelp)
    tkgrid(Save.but,Cancel.but,Defaults.but,Help.but,sticky="w")
    tkgrid(frame.but,sticky="n")
    
    tkfocus(tt)
    tkwait.window(tt)
    invisible()
 }

#############################################################################
#                                                                           #
#                               Important functions                         #
#                                                                           #
#############################################################################

# Calculates Hf initials ****** 
hfInitial<-function(age,x=WR){
    R<-x[,"176Lu/177Hf",drop=FALSE]
    I<-x[,"176Hf/177Hf",drop=FALSE]
    
    if(length(R)==1) age<-rep(age,length(R))
    y<-I-(R*(exp(lambda[1,"Hf"]*age*10^6)-1))
    age[is.na(age)]<--1
    y[age==0]<-I
    age[age==-1]<-NA
    return(y)
}

# Calculates initial Eps Hf values ****** 
hfEpsilon<-function(age,x=WR,digits=2){
    if(any(colnames(x)=="176Hf/177Hfi")){
        ee<-x[,"176Hf/177Hfi",drop=FALSE]
    }else{   
        ee<-hfInitial(age,x)
    }
    X<-round((ee/as.vector(hfInitial(age,CHUR[1,,drop=FALSE]))-1)*10^4,digits)
    
    if(is.matrix(X)){ 
        if(length(unique(age))==1){
            colnames(X)<-paste("EpsHf",unique(age),sep="_")
        }else{
            colnames(X)<-paste("EpsHf","i",sep="_")
        }
    }
    return(X)
}

# Calculates single-stage DM model ages ****** 
hfDMage<-function(x=WR,digits=3){
    R<-x[,"176Lu/177Hf",drop=FALSE]
    I<-x[,"176Hf/177Hf",drop=FALSE]
    X<-round(1/lambda[1,"Hf"]*log(((I-DM[1,"176Hf/177Hf"])/(R-DM[1,"176Lu/177Hf"]))+1)/10^9,digits)
    
    if(is.matrix(X)) colnames(X)<-"HfTDM.1stg"
    
    return(X)
}

# Calculates single-stage CHUR model ages ****** 
hfCHURage<-function(x=WR,digits=3){
    R<-x[,"176Lu/177Hf",drop=FALSE]
    I<-x[,"176Hf/177Hf",drop=FALSE]
    X<-round(1/lambda[1,"Hf"]*log(((I-CHUR[1,"176Hf/177Hf"])/(R-CHUR[1,"176Lu/177Hf"]))+1)/10^9,3)

    if(is.matrix(X)) colnames(X)<-"HfTCHUR.1stg"
    
    return(X)
}

# Calculates two-stage DM model ages
hfDM2stgAge<-function(x=WR,age=NULL,RCC=R.crust,digits=3){
    on.exit(options("show.error.messages"=TRUE))
    RCC<-as.numeric(RCC)

    # Age
    if(is.null(age)){
        options(show.error.messages=FALSE)        
        age<-try(x[,"Age",drop=FALSE])
        if(class(age)=="try-error"){cat("Invalid age.\n");stop(" ",call. = FALSE)}
        options(show.error.messages=TRUE)     
    }
    if(length(age)==1) age<-rep(age,nrow(x))
    
    # WR params
    R<-x[,"176Lu/177Hf",drop=FALSE]
    
    # Traditional approach
    I<-x[,"176Hf/177Hf",drop=FALSE]
    citatel<-I-(exp(lambda[1,"Hf"]*age*1e6)-1)*(R-RCC)-DM[,"176Hf/177Hf"]
    jmenovatel<-RCC-DM[,"176Lu/177Hf"]
    X<-round(1/lambda[1,"Hf"]*log((citatel)/(jmenovatel)+1)/10^9,digits)
    
    # Single-stage model age from the point t
    #I<-hfInitial(age,x)
    #X<-1/lambda[1,"Hf"]*log(((I-hfInitial(age,DM))/(RCC-DM[1,"176Lu/177Hf"]))+1)
    
    # Real age is to be converted to Ga and added to the single-stage TDM 
    #X<-round(X/10^9+age/1e3,digits) 
    
    if(is.matrix(X)) colnames(X)<-"HfTDM.2stg"

    return(X)
}


#############################################################################
#                                                                           #
#                                  MAIN LOOP                                #
#                                                                           #
#############################################################################

hfIso<-function(age=NULL, GUI=FALSE){
    if(length(grep("^176Hf/177Hf",colnames(WR)))==0)return()
    on.exit(options("show.error.messages"=TRUE))
    
    # Is sample name given to collate zircon analyses from each of them into coherent groups?
    prd<-which(toupper(colnames(labels))=="SAMPLE")
    if(length(prd)>0){ # Yes
        sample<-labels[,prd[1]]
        ee<-groupsByLabel(prd[1])
    }else{
        sample<-rep(NA,nrow(labels))
    }
    
    if(is.null(age)){ # No argument 'age' given upon the call. Provide one
        if(GUI|!any(toupper(colnames(WR))=="AGE")){
            age<-winDialogString("Age (Ma)",as.character(340))
            if(is.null(age)){
                cat("Cancelled.\n")
                options("show.error.messages"=FALSE)
                stop()
            }
            age<-as.numeric(age)
            age<-rep(age,times=nrow(WR))
        }else{
            age<-WR[,toupper(colnames(WR))=="AGE"]
        }
    }else{ # 'age' given upon the call, is the same for the whole dataset
        age<-rep(age,times=nrow(WR))
    }        
    
    if(any(colnames(WR)=="176Hf/177Hfi")){ # Use initial Hf isotopic ratios if present in the original file
        Hfinit<-WR[,"176Hf/177Hfi"]
    }else{   
        Hfinit<-hfInitial(age,WR)
    }
   
    eps<-hfEpsilon(age,WR)
    TDM<-hfDMage(WR)
    TCHUR<-hfCHURage(WR)
    TDM2stg<-hfDM2stgAge(WR,age) # Use the two-stage age
     
    Hfinit<-data.frame(age,as.character(sample),
        round(Hfinit,6),eps,TCHUR,TDM,TDM2stg,
        check.names=FALSE)
    
    rownames(Hfinit)<-rownames(WR)
    colnames(Hfinit)<-c("Age (Ma)","Sample","176Hf/177Hfi","EpsHfi","HfTCHUR.1stg","HfTDM.1stg","HfTDM.2stg")
    ee<-filterOut(Hfinit,colnames(Hfinit),n=5)
    if(!getOption("gcd.shut.up")){cat("\n");print(ee)}
    
    # Set up labels for plotting with correct formatting
    ii<-unique(age)
    if(length(ii)==1){ # Just a single age present
        b<-paste("epsilon[",ii,"]^Hf",sep="")
        b<-as.expression(b)
        epsHflabi<<-parse(text=b)
        
        b<-paste("\" \"^176*Hf/\" \"^177*Hf[",ii,"]",sep="")
        b<-as.expression(b)
        hflab<<-parse(text=b)
    }else{
        epsHflabi<<-expression(epsilon[i]^Hf)
        hflab<<-expression(" "^176*Hf/" "^177*Hf[i])
    }
    epsHflab<<-expression(epsilon[Hf])
    age<<-age
    results<<-Hfinit
    return(Hfinit)
}
Hfinit<-hfIso()
print(Hfinit)


#############################################################################
#                                                                           #
#                         Age vs EpsNdi GENERAL ENTRY POINT                 #
#                                                                           #
#############################################################################

# Internal function to overplot violins on Age-EpsHf plots
.overplotViolin<-function(which.samples = names(x.data),col = NULL){
    require("vioplot")
    names(groups)<-rownames(WR)
    fact<-factor(groups[which.samples],ordered=TRUE)
    
    # Obtain median ages for each of the groups
    med.ages<-tapply(x.data[which.samples],fact,median,na.rm=TRUE)

    # Assign the colours for the violin outlines
    if(is.null(col)){
        cols<-sheet$demo$call$col
        names(cols)<-which.samples

        col<-tapply(cols,fact,function(i){
            z<-col2rgb(i[1], alpha = FALSE)
            z<-rgb(t(z),alpha=66,maxColorValue = 255)
            return(z)
        })
    }
    vioplot(y.data[which.samples]~fact, at = med.ages, col = "transparent", border = col, lwd = 3,
        rectCol = "#00000066", lineCol = "#00000066", colMed = "white", pchMed = 18,        
        xaxt="n",yaxt="n", frame.plot=FALSE, main="", areaEqual = TRUE, wex=0.25, 
        add=TRUE)
}

# Age vs EpsNdi 1 stg ######################################################################
hfAgeEps<-function(){    
    on.exit(options("show.error.messages"=TRUE))
    ee<-select.list(c("EpsHf single-stage", "EpsHf two-stage","176Hf/177Hf two-stage"),
        preselect="EpsHf two-stage",multiple=FALSE) 
    if(nchar(ee)==0){cat("Cancelled.\n");options("show.error.messages"=FALSE);stop()}
      
    if(ee=="EpsHf single-stage"){
        ee<-hfAgeEps1(NULL);return()
    }

    if(ee=="EpsHf two-stage"){
        ee<-hfAgeEps2(NULL);return()
    }
    
    if(ee=="176Hf/177Hf two-stage"){
        ee<-hfAgeHfHf2(NULL);return()
    }
    invisible()
}

# Age vs EpsNdi 1 stg ######################################################################
hfAgeEps1<-function(which = rownames(Hfinit), 
    xmin = 0, xmax = NULL, ymin = NULL, ymax = 20,
    evol.lines1 = FALSE, evol.lines2 = TRUE, 
    violins = FALSE,
    age.hist = FALSE, kde.bandwidth = 30, rugs = TRUE, 
    pch = labels[which,"Symbol"], col = labels[which,"Colour"], 
    cex = labels[which, "Size"], new = TRUE,
    main = "Single-stage Hf isotopic growth diagram",...){
    
    if(is.null(which)){
        GUI<-TRUE
        which<-selectSubset(save=FALSE,GUI=TRUE)
    }else{
        GUI<-FALSE
    }    

    # Setting options if on GUI
    if(GUI){
        # y plotting range
        if(is.null(ymin)){
            y.present<-hfEpsilon(0,WR[rownames(Hfinit),c("176Lu/177Hf","176Hf/177Hf")])
            ymin<-min(pretty(y.present),na.rm=TRUE)
        }
        yrange <- winDialogString("Enter range for the y axis",paste(ymin,ymax,sep=","))
        if (is.null(yrange)) {
            cat("Cancelled.\n")
            options(show.error.messages = FALSE)
            stop(" ", call. = FALSE)
        }
        
        yy <- as.numeric(unlist(strsplit(yrange, ",")[[1]]))
        if (length(yy) != 2 | any(is.na(yy))) {
            options(show.error.messages = FALSE)
            winDialog("ok", "Invalid range!")
            cat(yrange, "not found.\n")
            stop(" ", call. = FALSE)
        }
        ymin<-yy[1]
        ymax<-yy[2]
        
        # First segment
        ee<-winDialog("yesno","Plot the first segments of the evolution lines?")
        evol.lines1<-ee=="YES"
        
        # Second segment
        ee<-winDialog("yesno","Plot the second segments of the evolution lines?")
        evol.lines2<-ee=="YES"

        # Rugs of model ages and initial epsilon Hf values
        ee<-winDialog("yesno","Add rugs portraying model ages and initial epsilon Hf values?")
        rugs<-ee=="YES"
            
        # Histogram of ages
        ee<-winDialog("yesno","Add frequency histogram of U-Pb ages?")
        age.hist<-ee=="YES"
    }

    # Plotting attributes
    if(length(pch)==1&length(which)>1){
        pch<-rep(pch,length(which))
    }
    names(pch)<-which
    
    if(length(col)==1&length(which)>1){
        col<-rep(col,length(which))
    }
    names(col)<-which
    
    if(length(cex)==1&length(which)>1){
        cex<-rep(cex,length(which))
    }
    names(cex)<-which
    
    # Plotting coordinates
    x<-age[which]/1000
    names(x)<-rownames(Hfinit[which,])
    y<-Hfinit[which,"EpsHfi"]
    names(y)<-rownames(Hfinit[which,])
    
    # x plotting scale
    if(is.null(xmax)){
        xmax<-max(pretty(Hfinit[,"HfTDM.1stg"]),na.rm=TRUE)
    }
   
    # y plotting scale
    if(is.null(ymin)){
        y.present<-hfEpsilon(0,WR[rownames(Hfinit),c("176Lu/177Hf","176Hf/177Hf")])
        ymin<-min(pretty(y.present),na.rm=TRUE)
    }
    
    # Plot main diagram
    plotWithLimits(x,y,xmin=xmin,ymax=ymax,ymin=ymin,xmax=xmax,
        xlab="Age (Ga)",ylab=epsHflab,col=col,pch=pch,cex=cex,new=new,main=main,...)
   
    # Hf isotopic development lines for our samples
    temp.lines<-list()
    if(evol.lines1&evol.lines2){ # To make sure it is a single line
          uf<-sapply(1:length(which),function(i){  
            if(!is.na(Hfinit[which[i],"176Hf/177Hfi"])){         
                x<-c(0,Hfinit[which[i],"HfTDM.1stg"])
                y<-c(hfEpsilon(0,WR[which[i],,drop=FALSE]),hfEpsilon(Hfinit[which[i],"HfTDM.1stg"]*10^3,DM))

                ee<-eval(parse(text=paste("temp.lines$lines",i,"A<<-list(\"lines\",x=c(", 
                    paste(x,collapse=","),"),y=c(", paste(y,collapse=","),"),
                    col= \"",col[which[i]],"\",lty=\"solid\")",sep="")))
                }
          },simplify=TRUE)
    }
    
    if(evol.lines1&!evol.lines2){ # Just first segment
          uf<-sapply(1:length(which),function(i){  
            if(!is.na(Hfinit[which[i],"176Hf/177Hfi"])){         
                x<-c(0,Hfinit[which[i],"Age (Ma)"]/1e3)
                y<-c(hfEpsilon(0,WR[which[i],,drop=FALSE]),Hfinit[which[i],"EpsHfi"])

                ee<-eval(parse(text=paste("temp.lines$lines",i,"A<<-list(\"lines\",x=c(", 
                    paste(x,collapse=","),"),y=c(", paste(y,collapse=","),"),
                    col= \"",col[which[i]],"\",lty=\"solid\")",sep="")))
                }
          },simplify=TRUE)
    }
    
    if(!evol.lines1&evol.lines2){ # Just second segment
          uf<-sapply(1:length(which),function(i){  
            if(!is.na(Hfinit[which[i],"176Hf/177Hfi"])){
                x<-c(Hfinit[which[i],"Age (Ma)"]/1e3,Hfinit[which[i],"HfTDM.1stg"])
                y<-c(Hfinit[which[i],"EpsHfi"],hfEpsilon(Hfinit[which[i],"HfTDM.1stg"]*10^3,DM))

                ee<-eval(parse(text=paste("temp.lines$lines",i,"B<<-list(\"lines\",x=c(", 
                    paste(x,collapse=","),"),y=c(", paste(y,collapse=","),"),
                    col= \"",col[which[i]],"\",lty=\"solid\")",sep="")))
                }
          },simplify=TRUE)
    }
    
    sheet$demo$template<-c(sheet$demo$template,temp.lines)
    
    sheet$demo$call$col<-col 
    sheet$demo$call$pch<-pch
    sheet$demo$call$cex<-cex
    
    # CHUR
    sheet$demo$template$abline1<-list("abline",h=0,col="darkgrey",lwd=3)
    Q<--CHUR[1,"176Lu/177Hf"]*lambda[1,"Hf"]/CHUR[1,"176Hf/177Hf"]*1e4*1e9
    
    # DM
    sheet$demo$template$abline2<-list("abline",a=(DM[1,"176Hf/177Hf"]/CHUR[1,"176Hf/177Hf"]-1)*1e4,
        b=Q*(DM[1,"176Lu/177Hf"]-CHUR[1,"176Lu/177Hf"])/CHUR[1,"176Lu/177Hf"],col="darkgrey",lwd=3)
    
    # Text labels for CHUR, DM
    xmax<-sheet$demo$call$xlim[2]
    sheet$demo$template$text1<-list("text",x=xmax-xmax/20,y=0,text="CHUR",size=1.5,pos=1)
    sheet$demo$template$text2<-list("text",x=xmax-xmax/20,y=15-1/.456*xmax-1,text="DM",size=1.5)
    
    # Rugs with initial epsilon values and model ages
    if(rugs){
        sheet$demo$template$rug1<-list("rug",x=Hfinit[which,"HfTDM.1stg"],ticksize=0.03,side=1,lwd=1,col="darkred")
        sheet$demo$template$rug2<-list("rug",x=Hfinit[which,"EpsHfi"],ticksize=0.03,side=2,lwd=1,col="darkred")  
    }
    
    # Frequency histogram of ages
    if(age.hist) sheet$demo$template$polygon1<-.HfFrequencyHistAgesOverplot(rugs=rugs,kde.bandwidth = kde.bandwidth)
    
    assign("sheet",sheet,.GlobalEnv)
    pp<<-figaro(demo,prefix="sheet")
    assign("pp",pp,.GlobalEnv)
    figRedraw()
    if(new){
        figaroOn()
    }
    
    if(violins) .overplotViolin()
    
    #if(GUI){
    #    .hfContour(GUI)
    #}
    
    # Plotting on a plate?
    ii<-screen()
    if(ii){
        plate[[ii]]<-sheet
        assign("plate",plate,.GlobalEnv)
        assign("x.data",plate.data[[ii]]$x,.GlobalEnv)
        assign("y.data",plate.data[[ii]]$y,.GlobalEnv)
        figRedraw()
    }    
    
    invisible()
}

# Age vs 176Hf/177Hfi 2 stg ######################################################################
hfAgeHfHf2<-function(which = rownames(Hfinit), RCC = R.crust, 
    xmin = 0, xmax = NULL, ymin = NULL, ymax = DM[,"176Hf/177Hf"],  
    evol.lines1 = FALSE, evol.lines2 = TRUE, violins = FALSE,
    age.hist = FALSE, kde.bandwidth = 30, rugs = TRUE,
    pch = labels[which,"Symbol"], col = labels[which,"Colour"], 
    cex = labels[which, "Size"], new = TRUE,
    main = "Two-stage Hf isotopic growth diagram",...){
    
    if(is.null(which)){
        GUI<-TRUE
        which<-selectSubset(save=FALSE,GUI=TRUE)
    }else{
        GUI<-FALSE
    }    

    # Setting options if on GUI
    if(GUI){
        # y plotting range
        if(is.null(ymin)){
            y.present<-WR[rownames(Hfinit),"176Hf/177Hf"]
            ymin<-min(pretty(y.present),na.rm=TRUE)
        }
        yrange <- winDialogString("Enter range for the y axis",paste(ymin,ymax,sep=","))
        if (is.null(yrange)) {
            cat("Cancelled.\n")
            options(show.error.messages = FALSE)
            stop(" ", call. = FALSE)
        }
        
        yy <- as.numeric(unlist(strsplit(yrange, ",")[[1]]))
        if (length(yy) != 2 | any(is.na(yy))) {
            options(show.error.messages = FALSE)
            winDialog("ok", "Invalid range!")
            cat(yrange, "not found.\n")
            stop(" ", call. = FALSE)
        }
        ymin<-yy[1]
        ymax<-yy[2]
        
        # Slope of the second segment
        msg<-paste("176Lu/177Hf of RCC (second segment)\nDefault for ",rownames(RCC),": ",RCC,sep="")
        rcc<- winDialogString(msg,as.character(RCC))
        
        if (is.null(rcc)) {
            cat("Cancelled.\n")
            options(show.error.messages = FALSE)
            stop(" ", call. = FALSE)
        }
        
        if(rcc!=RCC){
            prd <- as.numeric(rcc)
            prd <- data.matrix(prd)
            rownames(prd)<-"(User-defined RCC)"
            colnames(prd)<-colnames(RCC)
            RCC<-prd
        }
        
        if (is.na(RCC)){
            options(show.error.messages = FALSE)
            winDialog("ok", "Invalid range!")
            cat(yrange, "not found.\n")
            stop(" ", call. = FALSE)
        }
        
        # First segment
        ee<-winDialog("yesno","Plot the first segments of the evolution lines?")
        evol.lines1<-ee=="YES"

        # Second segment
        ee<-winDialog("yesno","Plot the second segments of the evolution lines?")
        evol.lines2<-ee=="YES"
        
        # Rugs of model ages and initial epsilon Hf values
        ee<-winDialog("yesno","Add rugs portraying model ages and initial epsilon Hf values?")
        rugs<-ee=="YES"
            
        # Histogram of ages
        ee<-winDialog("yesno","Add frequency histogram of U-Pb ages?")
        age.hist<-ee=="YES"
    }
    
    # Plotting attributes
    if(length(pch)==1&length(which)>1){
        pch<-rep(pch,length(which))
    }
    names(pch)<-which
    
    if(length(col)==1&length(which)>1){
        col<-rep(col,length(which))
    }
    names(col)<-which
    
    if(length(cex)==1&length(which)>1){
        cex<-rep(cex,length(which))
    }
    names(cex)<-which
    
    # Plotting coordinates
    x<-age[which]/1000
    names(x)<-rownames(Hfinit[which,])
    y<-Hfinit[which,"176Hf/177Hfi"]
    names(y)<-rownames(Hfinit[which,])
    
    # x plotting scale
    if(is.null(xmax)){
        xmax<-max(pretty(Hfinit[,"HfTDM.2stg"]),na.rm=TRUE)
    }
   
    # y plotting scale
    if(is.null(ymin)){
        y.present<-WR[rownames(Hfinit),"176Hf/177Hf"]
        ymin<-min(pretty(y.present),na.rm=TRUE)
    }
    
    # Plot main diagram
    plotWithLimits(x,y,xmin=xmin,ymax=ymax,ymin=ymin,xmax=xmax,
        xlab="Age (Ga)",ylab=annotate("176Hf/177Hf"),col=col,pch=pch,cex=cex,new=new,main=main,...)
 
    # Hf isotopic development lines for our samples
    temp.lines<-list()
    if(evol.lines1&evol.lines2){ # To make sure it is a single line
        uf<-sapply(1:length(which),function(i){  
            if(!is.na(Hfinit[which[i],"176Hf/177Hfi"])){         
                x.prc<-hfDM2stgAge(WR[which[i],,drop=FALSE],age=Hfinit[which[i],"Age (Ma)"],RCC=RCC,digits=10)
                xx<-c(0,Hfinit[which[i],"Age (Ma)"]/1e3,x.prc) # Age Ga
               
                y.prc<-cbind(RCC,Hfinit[which[i],"176Hf/177Hfi",drop=FALSE])
                colnames(y.prc)<-c("176Lu/177Hf","176Hf/177Hf")
                
                yy<-c(
                    WR[which[i],"176Hf/177Hf",drop=FALSE],
                    Hfinit[which[i],"176Hf/177Hfi"],
                    hfInitial(x.prc*1e3-Hfinit[which[i],"Age (Ma)"],y.prc)[,1] 
                )
                
                ee<-eval(parse(text=paste("temp.lines$lines",i,"A<<-list(\"lines\",x=c(", 
                    paste(xx,collapse=","),"),y=c(", paste(yy,collapse=","),"),
                    col= \"",col[which[i]],"\",lty=\"solid\")",sep="")))
                }
          },simplify=TRUE)
    }
     
    if(evol.lines1&!evol.lines2){ # Just the first segment
          uf<-sapply(1:length(which),function(i){  
            if(!is.na(Hfinit[which[i],"176Hf/177Hfi"])){         
                xx<-c(0,Hfinit[which[i],"Age (Ma)"]/1e3) # Age Ga
                yy<-c(WR[which[i],"176Hf/177Hf",drop=FALSE],Hfinit[which[i],"176Hf/177Hfi"])
                
                ee<-eval(parse(text=paste("temp.lines$lines",i,"A<<-list(\"lines\",x=c(", 
                    paste(xx,collapse=","),"),y=c(", paste(yy,collapse=","),"),
                    col= \"",col[which[i]],"\",lty=\"solid\")",sep="")))
                }
          },simplify=TRUE)
    }
 
    #x.prc<-hfDM2stgAge(WR[which,,drop=FALSE],age=Hfinit[which,"Age (Ma)"],RCC=RCC,digits=10)
    if(!evol.lines1&evol.lines2){ # Just the second segment  
          uf<-sapply(1:length(which),function(i){  
            if(!is.na(Hfinit[which[i],"176Hf/177Hfi"])){
                x.prc<-hfDM2stgAge(WR[which[i],,drop=FALSE],age=Hfinit[which[i],"Age (Ma)"],RCC=RCC,digits=10)
                xx<-c(Hfinit[which[i],"Age (Ma)"]/1e3,x.prc)
                
                y.prc<-cbind(RCC,Hfinit[which[i],"176Hf/177Hfi",drop=FALSE])
                colnames(y.prc)<-c("176Lu/177Hf","176Hf/177Hf")
                
                yy<-cbind(Hfinit[which[i],"176Hf/177Hfi"],
                    hfInitial(x.prc*1e3-Hfinit[which[i],"Age (Ma)"],y.prc) 
                )

                ee<-eval(parse(text=paste("temp.lines$lines",i,"B<<-list(\"lines\",x=c(", 
                    paste(xx,collapse=","),"),y=c(", paste(yy,collapse=","),"),
                    col= \"",col[which[i]],"\",lty=\"solid\")",sep="")))
                }
          },simplify=TRUE)          
    }
    
    if(evol.lines1&!evol.lines2){ # Just the first segment
        # Display vector of RCC
        deltax<-diff(par("usr")[1:2])/15
        
        x1<-par("usr")[2]-3*deltax
        x0<-x1-3*deltax
        
        # Mean 176Hf/177Hf of the whole dataset
        y0<-mean(y,na.rm=TRUE)
        y0HfHf<-cbind(RCC,y0)
        colnames(y0HfHf)<-c("176Lu/177Hf","176Hf/177Hf")
        
        # 176Hf/177Hf at x1
        y1<-hfInitial((x1-x0)*1e3,y0HfHf)
        
        sheet$demo$template$vector1<-list("lines",x=c(x0,x1),y=c(y0,y1),lwd=3,col="khaki4")
        
        # Text label for the vector
        srt<-atan((y1-y0)/diff(par("usr")[3:4])*diff(par("usr")[1:2])/(x1-x0))/pi*180 
        
        expr1<-parse(text=paste("\"\"^176*Lu/\"\"^177*Hf*\" = \"*",RCC,sep=""))
        sheet$demo$template$vector2<-list("text",x=mean(c(x0,x1)),y=mean(c(y0,y1)),text=expr1,srt=srt,pos=3,adj=0.5,cex=1,col="khaki4")
        
        expr2<-rownames(RCC)
        expr2<-gsub(" (","\n(",expr2,fixed=TRUE)
        if(length(expr2)==0) expr2<-"(User-defined RCC)"
        
        sheet$demo$template$vector3<-list("text",x=mean(c(x0,x1)),y=mean(c(y0,y1)),text=expr2,srt=srt,pos=1,adj=0.5,cex=0.8,col="khaki4")
   }
    
    sheet$demo$template<-c(sheet$demo$template,temp.lines)
    
    sheet$demo$call$col<-col 
    sheet$demo$call$pch<-pch
    sheet$demo$call$cex<-cex
    
    if(is.null(rownames(RCC))){
        expr1<-parse(text=paste("\"User-defined RCC with \"^176*Lu/\"\"^177*Hf*\" = \"*",RCC,sep=""))
        sheet$demo$call$sub<-expr1
    }else{
        sheet$demo$call$sub<-rownames(RCC)
    }
    
    # CHUR
    sheet$demo$template$churline<-list("lines",x=par("usr")[1:2],y=hfInitial(par("usr")[1:2]*1e3,CHUR),col="darkgrey",lwd=3)
    
    # DM
    sheet$demo$template$dmline<-list("lines",x=par("usr")[1:2],y=hfInitial(par("usr")[1:2]*1e3,DM),col="darkgrey",lwd=3)
        
    # Text labels for CHUR, DM
    xxx<-par("usr")[1]+diff(par("usr")[1:2])/15
    
    sheet$demo$template$churtext<-list("text",x=xxx,y=hfInitial(xxx*1e3,CHUR),text="CHUR",size=1.5,pos=1,adj=1)
    sheet$demo$template$dmtext<-list("text",x=xxx,y=hfInitial(xxx*1e3,DM),text="DM",size=1.5,pos=3,adj=1)
    
    # Rugs with initial epsilon values and model ages
    if(rugs){
        x.prc<-hfDM2stgAge(WR[which,,drop=FALSE],age=Hfinit[which,"Age (Ma)"],RCC=RCC,digits=10)
       
        sheet$demo$template$rug1<-list("rug",x=x.prc,ticksize=0.03,side=1,lwd=1,col="darkred")
        sheet$demo$template$rug2<-list("rug",x=Hfinit[which,"176Hf/177Hfi"],ticksize=0.03,side=2,lwd=1,col="darkred")  
    }
    
    # Frequency histogram of ages
    if(age.hist) sheet$demo$template$polygon1<-.HfFrequencyHistAgesOverplot(WR[which,"Age"],rugs=rugs, kde.bandwidth = kde.bandwidth)    

    assign("sheet",sheet,.GlobalEnv)
    pp<<-figaro(demo,prefix="sheet")
    assign("pp",pp,.GlobalEnv)
    
    # Broader left margin
    par("mar"=c(5.1,4.4,4.1,2.1))
    figRedraw()
    if(new){
        figaroOn()
    }
    
    if(violins) .overplotViolin()
    
    #if(GUI){
    #    .hfContour()
    #}
    
    # Plotting on a plate?
    ii<-screen()
    if(ii){
        plate[[ii]]<-sheet
        assign("plate",plate,.GlobalEnv)
        assign("x.data",plate.data[[ii]]$x,.GlobalEnv)
        assign("y.data",plate.data[[ii]]$y,.GlobalEnv)
        figRedraw()
    }    
    invisible()
}

# Age vs EpsNdi 2 stg ######################################################################
hfAgeEps2<-function(which = rownames(Hfinit), RCC = R.crust, 
    xmin = 0, xmax = NULL, ymin = NULL, ymax = 20,  
    evol.lines1 = FALSE, evol.lines2 = TRUE, violins = FALSE,
    age.hist = FALSE, kde.bandwidth = 30, rugs = TRUE,
    pch = labels[which,"Symbol"], col = labels[which,"Colour"], 
    cex = labels[which, "Size"], new = TRUE,
    main = "Two-stage Hf isotopic growth diagram",...){
    
    if(is.null(which)){
        GUI<-TRUE
        which<-selectSubset(save=FALSE,GUI=TRUE)
    }else{
        GUI<-FALSE
    }    

    # Setting options if on GUI
    if(GUI){
        # y plotting range
        if(is.null(ymin)){
            y.present<-hfEpsilon(0,WR[rownames(Hfinit),c("176Lu/177Hf","176Hf/177Hf")])
            ymin<-min(pretty(y.present),na.rm=TRUE)
        }
        yrange <- winDialogString("Enter range for the y axis",paste(ymin,ymax,sep=","))
        if (is.null(yrange)) {
            cat("Cancelled.\n")
            options(show.error.messages = FALSE)
            stop(" ", call. = FALSE)
        }
        
        yy <- as.numeric(unlist(strsplit(yrange, ",")[[1]]))
        if (length(yy) != 2 | any(is.na(yy))) {
            options(show.error.messages = FALSE)
            winDialog("ok", "Invalid range!")
            cat(yrange, "not found.\n")
            stop(" ", call. = FALSE)
        }
        ymin<-yy[1]
        ymax<-yy[2]
        
        # Slope of the second segment
        msg<-paste("176Lu/177Hf of RCC (second segment)\nDefault for ",rownames(RCC),": ",RCC,sep="")
        rcc<- winDialogString(msg,as.character(RCC))
        if (is.null(rcc)) {
            cat("Cancelled.\n")
            options(show.error.messages = FALSE)
            stop(" ", call. = FALSE)
        }
        
        if(rcc!=RCC){
            prd <- as.numeric(rcc)
            prd <- data.matrix(prd)
            colnames(prd)<-colnames(RCC)
            rownames(prd)<-"(User-defined RCC)"
            RCC<-prd
        }
        
        if (is.na(RCC)){
            options(show.error.messages = FALSE)
            winDialog("ok", "Invalid range!")
            cat(yrange, "not found.\n")
            stop(" ", call. = FALSE)
        }
        
        # First segment
        ee<-winDialog("yesno","Plot the first segments of the evolution lines?")
        evol.lines1<-ee=="YES"

        # Second segment
        ee<-winDialog("yesno","Plot the second segments of the evolution lines?")
        evol.lines2<-ee=="YES"
        
        # Rugs of model ages and initial epsilon Hf values
        ee<-winDialog("yesno","Add rugs portraying model ages and initial epsilon Hf values?")
        rugs<-ee=="YES"
            
        # Histogram of ages
        ee<-winDialog("yesno","Add frequency histogram of U-Pb ages?")
        age.hist<-ee=="YES"
    }
    
    # Plotting attributes
    if(length(pch)==1&length(which)>1){
        pch<-rep(pch,length(which))
    }
    names(pch)<-which
    
    if(length(col)==1&length(which)>1){
        col<-rep(col,length(which))
    }
    names(col)<-which
    
    if(length(cex)==1&length(which)>1){
        cex<-rep(cex,length(which))
    }
    names(cex)<-which
    
    # Plotting coordinates
    x<-age[which]/1000
    names(x)<-rownames(Hfinit[which,])
    y<-Hfinit[which,"EpsHfi"]
    names(y)<-rownames(Hfinit[which,])
    
    # x plotting scale
    if(is.null(xmax)){
        xmax<-max(pretty(Hfinit[,"HfTDM.2stg"]),na.rm=TRUE)
    }
   
    # y plotting scale
    if(is.null(ymin)){
        y.present<-hfEpsilon(0,WR[rownames(Hfinit),c("176Lu/177Hf","176Hf/177Hf")])
        ymin<-min(pretty(y.present),na.rm=TRUE)
    }
    
    # Plot main diagram
    plotWithLimits(x,y,xmin=xmin,ymax=ymax,ymin=ymin,xmax=xmax,
        xlab="Age (Ga)",ylab=epsHflab,col=col,pch=pch,cex=cex,new=new,main=main,...)
     
    # Hf isotopic development lines for our samples
    temp.lines<-list()
    if(evol.lines1&evol.lines2){ # To make sure it is a single line
         uf<-sapply(1:length(which),function(i){  
            if(!is.na(Hfinit[which[i],"176Hf/177Hfi"])){         
                x.prc<-hfDM2stgAge(WR[which[i],,drop=FALSE],age=Hfinit[which[i],"Age (Ma)"],RCC=RCC)
                xx<-c(0,Hfinit[which[i],"Age (Ma)"]/1e3,x.prc)
                
                y.prc<-cbind(RCC,Hfinit[which[i],"176Hf/177Hfi",drop=FALSE])
                colnames(y.prc)<-c("176Lu/177Hf","176Hf/177Hf")
                
                sample.init<-hfInitial(x.prc*1e3-Hfinit[which[i],"Age (Ma)"],y.prc)
                
                yy<-cbind(hfEpsilon(0,WR[which[i],,drop=FALSE]),
                    Hfinit[which[i],"EpsHfi"],
                    (sample.init/hfInitial(x.prc*1e3,CHUR)-1)*1e4
                ) 

                ee<-eval(parse(text=paste("temp.lines$lines",i,"A<<-list(\"lines\",x=c(", 
                    paste(xx,collapse=","),"),y=c(", paste(yy,collapse=","),"),
                    col= \"",col[which[i]],"\",lty=\"solid\")",sep="")))
                }
          },simplify=TRUE)
    }

    
    if(evol.lines1&!evol.lines2){ # Just the first segment
          uf<-sapply(1:length(which),function(i){  
            if(!is.na(Hfinit[which[i],"176Hf/177Hfi"])){         
                xx<-c(0,Hfinit[which[i],"Age (Ma)"]/1e3)
                yy<-c(hfEpsilon(0,WR[which[i],,drop=FALSE]),Hfinit[which[i],"EpsHfi"])

                ee<-eval(parse(text=paste("temp.lines$lines",i,"A<<-list(\"lines\",x=c(", 
                    paste(xx,collapse=","),"),y=c(", paste(yy,collapse=","),"),
                    col= \"",col[which[i]],"\",lty=\"solid\")",sep="")))
                }
          },simplify=TRUE)
    }

    if(!evol.lines1&evol.lines2){ # Just the second segment
          uf<-sapply(1:length(which),function(i){  
            if(!is.na(Hfinit[which[i],"176Hf/177Hfi"])){
                x.prc<-hfDM2stgAge(WR[which[i],,drop=FALSE],age=Hfinit[which[i],"Age (Ma)"],RCC=RCC)
                xx<-c(Hfinit[which[i],"Age (Ma)"]/1e3,x.prc)
                
                y.prc<-cbind(RCC,Hfinit[which[i],"176Hf/177Hfi",drop=FALSE])
                colnames(y.prc)<-c("176Lu/177Hf","176Hf/177Hf")
                
                sample.init<-hfInitial(x.prc*1e3-Hfinit[which[i],"Age (Ma)"],y.prc)
                
                yy<-cbind(Hfinit[which[i],"EpsHfi"],
                    (sample.init/hfInitial(x.prc*1e3,CHUR)-1)*1e4
                ) 
                
                ee<-eval(parse(text=paste("temp.lines$lines",i,"B<<-list(\"lines\",x=c(", 
                    paste(xx,collapse=","),"),y=c(", paste(yy,collapse=","),"),
                    col= \"",col[which[i]],"\",lty=\"solid\")",sep="")))
                }
          },simplify=TRUE)
    }
    
    if(evol.lines1&!evol.lines2){
        # Display vector of RCC
        deltax<-diff(par("usr")[1:2])/15
        
        x1<-par("usr")[2]-3*deltax
        x0<-x1-3*deltax
        y0<-0
        
        # CHUR at x0 Ma ago
        y0HfHf<-cbind(RCC,hfInitial(x0*1e3,CHUR[1,,drop=FALSE]))
        colnames(y0HfHf)<-c("176Lu/177Hf","176Hf/177Hf")
        
        # EpsHf at x1
        y1HfHf<-hfInitial((x1-x0)*1e3,y0HfHf)
        colnames(y1HfHf)<-"176Hf/177Hfi"
        y1HfHf<-cbind(RCC,y1HfHf)
        y1<-hfEpsilon(x1*1e3,y1HfHf)
        
        sheet$demo$template$vector1<-list("lines",x=c(x0,x1),y=c(y0,y1),lwd=3,col="khaki4")
        
        # Text label for the vector
        srt<-atan((y1-y0)/diff(par("usr")[3:4])*diff(par("usr")[1:2])/(x1-x0))/pi*180 
        
        expr1<-parse(text=paste("\"\"^176*Lu/\"\"^177*Hf*\" = \"*",RCC,sep=""))
        sheet$demo$template$vector2<-list("text",x=mean(c(x0,x1)),y=mean(c(y0,y1)),text=expr1,srt=srt,pos=3,adj=0.5,cex=1,col="khaki4")
        
        expr2<-rownames(RCC)
        expr2<-gsub(" (","\n(",expr2,fixed=TRUE)
        if(length(expr2)==0) expr2<-"(User-defined RCC)"
        
        sheet$demo$template$vector3<-list("text",x=mean(c(x0,x1)),y=mean(c(y0,y1)),text=expr2,srt=srt,pos=1,adj=0.5,cex=0.8,col="khaki4")
    }
    
    sheet$demo$template<-c(sheet$demo$template,temp.lines)
    
    sheet$demo$call$col<-col 
    sheet$demo$call$pch<-pch
    sheet$demo$call$cex<-cex
    
    if(is.null(rownames(RCC))){
        expr1<-parse(text=paste("\"User-defined RCC with \"^176*Lu/\"\"^177*Hf*\" = \"*",RCC,sep=""))
        sheet$demo$call$sub<-expr1
    }else{
        sheet$demo$call$sub<-rownames(RCC)
    }
    
    # CHUR
    sheet$demo$template$abline1<-list("abline",h=0,col="darkgrey",lwd=3)
    Q<--CHUR[1,"176Lu/177Hf"]*lambda[1,"Hf"]/CHUR[1,"176Hf/177Hf"]*1e4*1e9
    
    # DM
    sheet$demo$template$abline2<-list("abline",a=(DM[1,"176Hf/177Hf"]/CHUR[1,"176Hf/177Hf"]-1)*1e4,
        b=Q*(DM[1,"176Lu/177Hf"]-CHUR[1,"176Lu/177Hf"])/CHUR[1,"176Lu/177Hf"],col="darkgrey",lwd=3)
    
    # Text labels for CHUR, DM
    xmax<-sheet$demo$call$xlim[2]
    sheet$demo$template$text1<-list("text",x=xmax-xmax/20,y=0,text="CHUR",size=1.5,pos=1)
    sheet$demo$template$text2<-list("text",x=xmax-xmax/20,y=hfEpsilon((xmax-xmax/20)*1000,DM),
        text="DM",size=1.5,pos=3)
        
    # Rugs with initial epsilon values and model ages
    if(rugs){
        x.prc<-hfDM2stgAge(WR[which,,drop=FALSE],age=Hfinit[which,"Age (Ma)"],RCC=RCC)
        sheet$demo$template$rug1<-list("rug",x=x.prc,ticksize=0.03,side=1,lwd=1,col="darkred")
        sheet$demo$template$rug2<-list("rug",x=Hfinit[which,"EpsHfi"],ticksize=0.03,side=2,lwd=1,col="darkred")  
    }
    
    # Frequency histogram of ages
    if(age.hist) sheet$demo$template$polygon1<-.HfFrequencyHistAgesOverplot(WR[which,"Age"],rugs=rugs,kde.bandwidth = kde.bandwidth)    

    assign("sheet",sheet,.GlobalEnv)
    pp<<-figaro(demo,prefix="sheet")
    assign("pp",pp,.GlobalEnv)
    figRedraw()
    if(new){
        figaroOn()
    }
    
    if(violins) .overplotViolin()
    
    #if(GUI){
    #    .hfContour()
    #}
    
    
    # Plotting on a plate?
    ii<-screen()
    if(ii){
        plate[[ii]]<-sheet
        assign("plate",plate,.GlobalEnv)
        assign("x.data",plate.data[[ii]]$x,.GlobalEnv)
        assign("y.data",plate.data[[ii]]$y,.GlobalEnv)
        figRedraw()
    }    
    invisible()
}

.HfselectVariable<-function(){
    on.exit(options("show.error.messages"=TRUE))
    vars<-c("Age (Ma)","176Hf/177Hfi","EpsHfi","1stg CHUR model ages","1stg DM model ages","2stg DM model ages")
      
    what<-select.list(vars,preselect="EpsHfi",multiple=FALSE)
    if(nchar(what)==0){cat("Cancelled.\n");options("show.error.messages"=FALSE);stop()}
    
    ee<-c("Age (Ma)","176Hf/177Hfi","EpsHfi","HfTCHUR.1stg","HfTDM.1stg","HfTDM.2stg")[which(vars==what)]
    
    # Try it whether it does work
    options(show.error.messages=FALSE)
    prd<-try(Hfinit[,ee])
    if(class(prd)=="try-error"){
        cat("Variable ", prd," not found.\n",sep="");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)
    }
    return(c(ee,what))
}

.HfFrequencyHistAgesOverplot<-function(x=WR[,"Age"],col="lightblue",rugs=FALSE, kde.bandwidth = 30){
    hist(x, main = "", col = "transparent", border="transparent",
        prob = TRUE, axes=FALSE, labels = FALSE, plot=FALSE, breaks=nrow(WR)/6)
    dens<-density(x,bw=kde.bandwidth) 
    dify<-diff(par("usr")[3:4])
   
    scaley<-dify/max(dens$y,na.rm=TRUE)/5
    #  ee<-hfAgeEps2(evol.lines2=FALSE) # Debugging
    #figXlim(c(0.4,0.6)) # Debugging
    #figYlim(c(-60,16))  # Debugging
    
    # Extra space for rugs
    if(rugs) extra.sp<-dify/30 else extra.sp<-0
    
    out<-list("polygon",x=dens$x/1000,y=dens$y*scaley+par("usr")[3]+extra.sp, col=col, border="black")
    invisible(out)
}

.hfContour<-function(GUI=TRUE, levels = c(0.01, 0.1, 0.5), drawlabels = FALSE,...){
    if(GUI){
        # levels
        my.levels <- winDialogString("Enter the levels (0-1), separatged by commas,\nwhere individual contours are to be drawn",paste(levels,collapse=","))
        if (is.null(my.levels)) {
            cat("Cancelled.\n")
            options(show.error.messages = FALSE)
            stop(" ", call. = FALSE)
        }
        
        levels <- as.numeric(unlist(strsplit(my.levels, ",")[[1]]))
        if (any(is.na(levels))) {
            options(show.error.messages = FALSE)
            winDialog("ok", "Invalid range!")
            cat(my.levels, "not found.\n")
            stop(" ", call. = FALSE)
        }
        
        # Label each contour?
        ee<-winDialog("yesno","Label each of the contour lines?")
        drawlabels<-ee=="YES"
        
        # Colours
        col<-pickColour()
    }
    addContours(levels=levels, drawlabels = drawlabels,col=col,lty="solid",...)
    invisible()
}

#############################################################################
#                                                                           #
#             Histograms, boxplots, violin plots and stripplots             #
#                                                                           #
#############################################################################

hfHist<-function(what = NULL, x = NULL, freq = FALSE, 
    bin.number = NULL, bin.width = NULL, col = "lightblue", border = "black",
    kde = TRUE, kde.bandwidth = 30, ...){
    on.exit(options("show.error.messages"=TRUE))
    
    dotdot<-list(...)
    
    if(is.null(x) & is.null(what)){
        what<-.HfselectVariable()[1]
        GUI<-TRUE
    }else{
        GUI<-FALSE
    }
    
    windows(width = 5, height = 5, pointsize = 10,title=paste("Histogram of ",what)) # 
    
    # Selecting variable
    if(is.null(x)){
        options(show.error.messages=FALSE)
        x<-try(Hfinit[,what])
        if(class(x)=="try-error"){
        cat("Variable ", ee," not found.\n",sep="");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)
        }
    }
    
    # Graphical user interface
    if(GUI){
        # Number of bins
        xx<-winDialogString("Approximate number of bins","")
        if(is.null(xx)){cat("Cancelled.\n");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)}
        
        if(xx!=""){
            options(show.error.messages=FALSE)
            bin.number<-try(as.numeric(xx))
            if(class(bin.number)=="try-error"){
                cat("Invalid number of mins.\n");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)
            }
        }else{
            bin.number<-NULL
        }
        
        # Width of bins
        xx<-winDialogString("Enter width of a bin","")
        if(is.null(xx)){cat("Cancelled.\n");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)}
        
        if(xx!=""){
            options(show.error.messages=FALSE)
            bin.width<-try(as.numeric(xx))
            if(class(bin.width)=="try-error"){
                cat("Invalid number of mins.\n");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)
            }
        }else{
            bin.width<-NULL
        }
        
    }
    
    breaks<-"Sturges"
    
    # Setting number of bins
    if(!is.null(bin.number)){
        breaks<-bin.number   
    }
    
    # Setting width of a single bin
    if(!is.null(bin.width)){
        xxx<-pretty(x)
        breaks<-seq(min(xxx),max(xxx),by=bin.width)   
    }
    
    if(any(names(dotdot)=="breaks")){ 
        breaks<-dotdot$breaks
        dotdot$breaks<-NULL
    }    
   
    if(breaks!="Sturges"&length(breaks)==1&(breaks>length(x)|breaks<1)){
        cat("Invalid breaks value.\n");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)
    }
    
   # Histogram
   args<-c(list(x = x, main = "", xlab = annotate(what), col = col, border = border, 
        freq = freq, plot = TRUE, breaks = breaks), dotdot)
   out<-do.call("hist",args)
   box()
       
   # Kernel density curve if reqrd
   if(kde){
        if(GUI){
        # Number of bins
            kde.bandwidth<-winDialogString("Enter KDE bandwidth",as.character(kde.bandwidth))
            if(is.null(kde.bandwidth)){cat("Cancelled.\n");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)}
            if(kde.bandwidth==""){cat("Cancelled.\n");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)}
        }
        options(show.error.messages=FALSE)
        kde.bandwidth<-try(as.numeric(kde.bandwidth))
        if(class(kde.bandwidth)=="try-error"){
            cat("Invalid KDE bandwidth.\n");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)
        }
        
        dens<-density(x,n=length(x),bw=kde.bandwidth) 
        lines(dens,col="blue",lwd=1.5)
        #polygon(dens, col = "pink",lwd=1.5)
   }
    
   figaroOff()
   invisible(out)
}

hfBoxplot<-function(what=NULL,varwidth=TRUE,cex.axis=1,col="lightblue",horizontal=TRUE,xaxs="i",...){
    on.exit(options("show.error.messages"=TRUE))
    if(is.null(what)){
        what<-.HfselectVariable()#[1]
        GUI<-TRUE
    }else{
        GUI<-FALSE
    }
    options("show.error.messages"=TRUE)
    names(groups)<-rownames(WR)
    x<-factor(groups[rownames(WR)],ordered=TRUE)
    
    windows(width = 5, height = 5, pointsize = 10,title=paste("Boxplot of ",what)) # 
    par(las=1)
    par(oma=c(1,1,1,1))
    
    ee<-grep("model",what[2])
    if(length(ee)>0){
        xlab<-as.expression(parse(text=paste(annotate(what[1]),"*(Ga)",sep=""))) 
    }else{
        xlab<-annotate(what[1])
    }
    
    # Try whether it does work
    options(show.error.messages=FALSE)
    prd<-try(Hfinit[,what[1]])
    if(class(prd)=="try-error"){
        cat("Variable '", what,"' not found.\n",sep="");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)
    }
    
    # Boxplot
    ylim<-range(pretty(Hfinit[,what[1]],na.rm=TRUE))
    if(GUI){
        ylim<-winDialogString("Enter range for the numeric axis",paste(ylim,collapse=","))
        if(is.null(ylim)){cat("Cancelled.\n");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)}
        ylim<-as.numeric(unlist(strsplit(ylim,",")[[1]]))
    }
    
    out<-boxplot(Hfinit[,what[1]]~x,ylab="",xlab=xlab,varwidth=varwidth,cex.axis=cex.axis,
        col=col,horizontal=horizontal,xaxs=xaxs,main="",ylim=ylim,...) # main=annotate(what[2])
    rownames(out$stats)<-c("Min","25%","50%","75%","Max")
    colnames(out$stats)<-out$names
    names(out$n)<-out$names
    
    ee<-Hfinit[,what[1]]
    ee<-ee[!is.na(ee)]
    i<-pretty(ee)
    
    if(horizontal==TRUE){
        abline(v=i,col="gray",lty="dotted")
        abline(v=0,lty="dashed")
        abline(h=(0.5:35.5),lty="dotted",col="gray")
    }else{
        abline(h=i,col="gray",lty="dotted")
        abline(h=0,lty="dashed")
        abline(v=(0.5:35.5),lty="dotted",col="gray")
    }
    box()
    figaroOff()
    return(out)
}

hfViolinplot<-function(what = NULL, col = NULL, border = "black",
    rectCol = "white", lineCol = "white", pchMed = NULL, colMed = "black",  
    cex = 1.5, xlab = "", ylab = "", ...){ 
    # additonal useful args are axes, add, xlim, ylim, horizontal
    on.exit(options("show.error.messages"=TRUE))
    require("vioplot")
    
    if(is.null(what)){
        what<-.HfselectVariable()
        GUI<-TRUE
    }else{
        GUI<-FALSE
    }
    
    names(groups)<-rownames(WR)
    x<-factor(groups[rownames(WR)],ordered=TRUE)
    
    windows(width = 7, height = 5, pointsize = 10,title=paste("Violinplot of ",what))
    par(las=1)
    par(oma=c(1,1,1,1))
    
    ee<-grep("model",what[2])
    if(length(ee)>0){
        ylab<-as.expression(parse(text=paste(annotate(what[1]),"*(Ga)",sep="")))
        #ylab<-paste(what[2]," (Ga)",sep="")
    }else{
        ylab<-annotate(what[1])
    }
    
    # Assign the colours for the violin fills
    if(is.null(col)){
        col<-tapply(labels[,"Colour"],x,function(i){
            return(i[1])
        })
    }
    
    # Assign the plotting symbols for the median point
    if(is.null(pchMed)){
        pch<-tapply(labels[,"Symbol"],x,function(i){
            return(i[1])
        })
    }
    
    # Try whether it does work
    options(show.error.messages=FALSE)
    prd<-try(Hfinit[,what[1]])
    if(class(prd)=="try-error"){
        cat("Variable '", what,"' not found.\n",sep="");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)
    }

    # Violin plot
    ylim<-range(pretty(Hfinit[,what[1]],na.rm=TRUE))
    
    ylim<-range(pretty(Hfinit[,what[1]],na.rm=TRUE))
    if(GUI){
        ylim<-winDialogString("Enter range for the numeric axis",paste(ylim,collapse=","))
        if(is.null(ylim)){cat("Cancelled.\n");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)}
        ylim<-as.numeric(unlist(strsplit(ylim,",")[[1]]))
    }
    
    out<-vioplot(Hfinit[,what[1]]~x, main = ylab, col = col, 
        rectCol = rectCol, lineCol = lineCol, colMed = colMed, pchMed = pch, 
        ylim = ylim, xlab = xlab, ylab = " ", cex = cex, ...)
    figaroOff()
    return(out)
}

hfStripplot<-function(what=NULL){
    on.exit(options("show.error.messages"=TRUE))

    strip.main<-function(elem,xlab="",title="Stripplot"){    
        names(groups)<-rownames(WR)
        if(substr(elem,1,5)!="delta"){
            ee<-Hfinit[,elem,drop=FALSE]
        }else{
            ee<-WR[,elem,drop=FALSE]
        }
        ii<-!is.na(ee)
        which<-rownames(ee)[ii]
        
        ee<-ee[which,,drop=FALSE]
        x<-factor(groups[which],ordered=TRUE)
        trellis.device(bg = "white",new = TRUE,retain = FALSE,title=title)
        trellis.par.set("background",list(col="white"))
        stripplot(x ~ ee,aspect = 1,jitter = TRUE, pch=labels[which,"Symbol"],
            col=labels[which,"Colour"],xlab=xlab,cex=labels[which,"Size"])
    }
    
    if(is.null(what)){
        what<-.HfselectVariable()
        GUI<-TRUE
    }else{
        GUI<-FALSE
    }
    
    if(length(what)==1) what<-rep(what,2) 
    ee<-grep("model",what[2])
    
    if(length(ee)>0){
        xlab<-as.expression(parse(text=paste(annotate(what[1]),"*(Ga)",sep="")))
        #xlab<-paste(what[2]," (Ga)",sep="")
    }else{
        xlab<-annotate(what[1])
    }
    
    # Try whether it does work
    options(show.error.messages=FALSE)
    prd<-try(Hfinit[,what[1]])
    if(class(prd)=="try-error"){
        cat("Variable '", what[1],"' not found.\n",sep="");options("show.error.messages"=FALSE);stop(" ",call. = FALSE)
    }
    
    ee<-strip.main(what[1],xlab=xlab,title=paste("Stripplot of ",what[2],sep=""))
    print(ee)
    figaroOff()
}

hfLegend<-function(pos="topright",new.plot=FALSE,main=NULL,...){
    gr<-factor(groups)
    gr.counts<- tapply(groups,gr,length)
    
    if(new.plot){
        if(!is.null(dev.list())) dev.bak<-dev.cur() else dev.bak<-NULL
        windows(width = 3.5, height = 8.5,points=10)
    }
    
    labs<-as.character(levels(gr))
    labs[is.na(labs)]<-"Unspecified"
    xx<-.build.legend(labs=gr,chars=TRUE,cols=TRUE)
    xx[,1]<-paste(names(gr.counts),", n = ",gr.counts,sep="")
    if(length(unique(xx[,2]))==1 & length(unique(xx[,3]))==1){
        xx<-xx[1,,drop=FALSE]
        xx[1,1]<-"All the same"
    }   
    
    if(is.vector(xx))xx<-t(xx);rownames(xx)<-xx[,1]
    pch<-xx[,2]
    if(all(!is.na(as.numeric(pch)))){
    pch<-as.numeric(pch)
    
    col<-xx[,3]

    n<-length(labs)

    if(new.plot){
        plot(1,1,xlim=c(-1,1),ylim=c(-1,1),type="n",xlab="",ylab="",axes=FALSE)
        text(0,1,annotate(main),cex=1,adj=0.5)
    }
    labs<-rownames(xx)
    
    go<- as.go(points=as.pgo(rep(1,n),1:n,pch=pch,col=col,label=labs))
    
    if(new.plot){
        ee<-legend.go(go,xy=list(x=-1,y=0.9),...) 
        if(!is.null(dev.bak)) dev.set(dev.bak)
    }else{
        ee<-legend.go(go,xy=list(x=pos,y=NULL),bg="white",...)
        }
    }
    invisible()
}



#############################################################################
#                                                                           #
#                         Dealing with Hf isotopic results                  #
#                                                                           #
#############################################################################

hfAddResults<-function(){
    if(is.na(all( match(colnames(Hfinit),colnames(WR))))){
        cat("Appending...\n")
        addResults("Hfinit")
    }else{
        cat("Replacing...\n")
        ee<-WR
        ee[rownames(Hfinit),colnames(Hfinit)]<-Hfinit
        assign("WR",ee,.GlobalEnv)
    }    
}

.switch.results.export.format<-function(digits=6){
    on.exit(options("show.error.messages"=TRUE))
    if(R.version$arch=="x86_64"){
        where<-c("TXT","HTML")
    }else{
        where<-c("TXT","HTML","Excel (XLS)")#,"Excel 2007 (XLSX)","Access (MDB)","dBase (DBF)")
    }
    selected<-utils::select.list(where,title="Select output format",multiple=FALSE)
    if(is.null(selected))selected<-""
    if(selected==""){cat("Cancelled.\n");options(show.error.messages=FALSE);stop()}
    
    ee<-switch(which(where==selected),
        "saveResults(Hfinit,digits=digits)",
        "HTMLTableMain(Hfinit)",
        "excelExport(Hfinit,transpose=FALSE,dec.places=digits)"#,
        #"excel2007Export(Hfinit,transpose=FALSE,dec.places=digits)",
        #"accessExport(Hfinit,transpose=FALSE,dec.places=digits)",
        #"dbfExport(Hfinit)"    
    )
    cat("GCDkit->",ee,"\n")
    .save2hist(ee)
    eval(parse(text=ee)) 
}

winMenuDel("$ConsolePopup/Results table")
winMenuAdd("$ConsolePopup/Results table")
    winMenuAddItem("$ConsolePopup/Results table","Copy to clipboard","r2clipboard(Hfinit)")
    winMenuAddItem("$ConsolePopup/Results table","Append to data","addResults(,GUI=TRUE)")
    winMenuAddItem("$ConsolePopup/Results table","----------------------------------","none")
    winMenuAddItem("$ConsolePopup/Results table","Save to text file","saveResults(Hfinit,digits=6)")
    winMenuAddItem("$ConsolePopup/Results table","----------------------------------  ","none")
    winMenuAddItem("$ConsolePopup/Results table","Export to HTML","HTMLTableMain(Hfinit)")
    if(R.version$arch!="x86_64"){
        winMenuAddItem("$ConsolePopup/Results table","Export to Excel","excelExport(Hfinit,transpose=FALSE,dec.places=6)")
        #winMenuAddItem("$ConsolePopup/Results table","Export to Excel 2007","excel2007Export(Hfinit,transpose=FALSE,dec.places=6)")
        #winMenuAddItem("$ConsolePopup/Results table","Export to Access","accessExport(Hfinit,transpose=FALSE,dec.places=6)")
        #winMenuAddItem("$ConsolePopup/Results table","Export to DBF","dbfExport(Hfinit,transpose=FALSE)")
    }


hfSaveResults<-function(digits = 6){ # 
    .switch.results.export.format(digits)
}

# Adding items to the Plugin menu ##############################################################
.hfMenu<-function(){
    on.exit(options("show.error.messages"=TRUE)) 
    where<-c(
        "See/change calculation parameters", "Recalculate to new age", "Copy Hf isotopic data to clipboard",    
        "Save Hf isotopic data", "Append Hf isotopic data", 
        "------------------------------------    ",
        "Hafnium DM growth lines",
        "------------------------------------      ",
        "Histogram ages/isotopic ratios/model ages",
        "Boxplot ages/isotopic ratios/model ages",
        "Stripplot ages/isotopic ratios/model ages",
        "Violinplot ages/isotopic ratios/model ages",
        "------------------------------------        ",
        "Display legend"
    )
    
    selected<-select.list(where, multiple = FALSE, title = "Hafnium isotopes")
    if(is.null(selected)) selected<-""
    if(selected==""){cat("Cancelled.\n");options("show.error.messages"=FALSE);stop()}
    ee<-switch(which(where==selected),
        "hfOptions()",
        "Hfinit<<-hfIso(GUI=TRUE);print(Hfinit)", 
        "r2clipboard(Hfinit)", 
        "ee<-hfSaveResults(digits=6)", 
        "ee<-hfAddResults()", 
        "", 
        "ee<-hfAgeEps()", 
        "", 
        "hfHist()",
        "hfBoxplot()", 
        "hfStripplot()",
        "hfViolinplot()",
        "",
        "hfLegend(new.plot=TRUE)"
    )
    cat("GCDkit->",ee,"\n")
    if(.Platform$OS.type=="windows").save2hist(ee)
    eval(parse(text=ee))
} 

if(getOption("gcd.menus")!=""){winMenuAddItem("Plugins","Hf isotopes",".hfMenu()")}

}else{
    if(!getOption("gcd.shut.up"))cat("WARNING: Skipping: No Hf in Zrn isotopic data present!")
}
